# Changelog

Download sensor plugins from <https://hidale.com/shop/>  
Setup and install guide is in file `INSTALL.md`
This software comes with ABSOLUTELY NO WARRANTY.

## v1.3.20240219

* Fix slow memory leak with ml dictionary output
* Update to Onnxruntime v1.16.3 - memory improvements, caching intermediate models, more ONNX ops
* Optimize color transcoders
* Add `@mlinput` to specify the source and channels using a list of FourCC codes for the input
  tensors for your model. The default `@mlinput rgb` is a single color tensor with 3-channels in
  the order red green blue. Choose the FourCC codes needed for your model. `@mlinput` may be easier
  to understand than `@mlswapch`. For example:
  ```sh
  @mlinput grey     # greyscale, 1-channel, 8-bit integer
  @mlinput bgr      # color, 3-channel in order blue green red, type of @colortype
  @mlinput rgba     # color, 4-channel in order red green blue alpha, type of @colortype
  ```

## v1.3.20231220

Microsoft has discontinued the Kinect v3 (Azure Kinect DK).
[Orbbec](https://www.orbbec.com/) makes a depth sensor based on the same
Kinect v3 sensor technology. They claim their
[Femto Bolt](https://www.orbbec.com/products/tof-camera/femto-bolt/)
is very similar hardware but requires recompiling software to use it.
We are exploring the possibility to support this sensor in the future.

### Changes

* Data processing pipeline simplified, more flexible, less memory use.
* Add `@mount` to indicate how your sensor is mounted (upside up, upside down, left, right).
  Some models (like Azure Body Tracking) may use this to optimize their inference.
* `@decodercolor` is deprecated and replaced by `@transcoder`. `@decodercolor intelmedia` is
  still supported. This is simplier to use. The default is `@transcoder auto` which
  automatically accelerates transcoding/decoding. The only other option is
  `@transcoder intelmedia` which uses the Intel GPU hardware transcoder (also known as QuickSync)
  when possible and then falls back to the `auto` acceleration.
* Fix setting `@colormap` and some other attributes with symbolic values
* Renamed incorrect `@colorbalance` to `@whitebalance`.
* Update onnxruntime to v1.15.1 which fixs several bugs including model caching.
* `uyvy` is better supported...though still not recommended due to widespread Max v6, v7, and v8
  bugs processing and displaying `uyvy` in patches.
* Improved use of OpenCL device for image postprocessing
* Clearer error messages when fatal errors occur
* Fix rare frame desync and large memory use if CPU is consistently overloaded for long periods
* Smaller plugin file size

## v1.3.20221003

### Changes

* Upgrade to Kinect Body Tracking SDK v1.1.2. Microsoft adds crashing
  bugs (see below) while also fixing some memory leaks and compatibility issues.
* Add color white balance
* Add ear, eyes, nose joints and their orientation
* Add symbolic device name support to `@opencl`. The numeric index is still supported.
  The symbolic name is any case-insensitive substring of the full device name. For example:
  @opencl off           @opencl default
  @opencl intel         @opencl uhd
  @opencl nvidia        @opencl rtx
  @opencl "2070"  <--- notice the number is quoted as a string because numbers are devices indexes
* Support for Nvidia CUDA compute 3.0 (e.g. GeForce GTX 680) has been removed.
  These GPUs are ten years old and the SDKs themselves have removed support for them.
* Use custom ONNX runtime to fix onnx bugs -and- fix Microsoft bugs
  <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1607>,
  <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1510>
* Removed `gpu` choice from `@skelcompute`. Use `directml` instead.
* Add machine learning inference. Expect feature changes and bugs if you use this early feature.
  The ml inference is output on the 5th outlet as a max dictionary. (route) or (routepass) can
  be used to send it to other dict objects for post-processing. Models are restricted to those
  that have a single input that accepts RGB or BGR, and that output numeric data (not strings).
  The color image from the sensor is the source of RGB and is always scaled to the model input's
  size requirements *without* preserving aspect ratio; see `@mlcrop` and `@mlswapch` below.
  SqueezeNet <https://github.com/onnx/models/tree/main/vision/classification/squeezenet> is a
  model I use during development.
  * `@mlmodel` load an ONNX model file and enable the ml feature
  * `@mlcompute` is your choice of directml or cpu device; supports symbolic names
  * `@mlcrop` center crop input color image and keep aspect ratio
  * `@mlmean` values in channel 0,1,2 (RGB) order are *subtracted* from every pixel `(pixel - mean) * scale`
  * `@mlscale` is multiplied to every pixel `(pixel - mean) * scale`; often used for std deviation
  * `@mlswapch` applies `@mlmean` to channels 2,1,0 *and* swaps data channels 0 and 2
    before sending to the model as input; use when your model requires BGR input

### Known Issues

* Microsoft body tracking v1.1.2 SDK crashes when you close Max in some combinations of body
  tracking compute choices set with `@skelcompute`. You might not see the crash and it may
  be no real-world harm. <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1783>
* Max may freeze or crash with the below scenario. It may be safer to change
  `@skelcompute` by typing that attribute on the (dp.kinect3) node.
  1. Output visual data like depth, color, ir, or playermap
  2. Stop dp.kinect3
  3. Change `@skelcompute` using a message to dp.kinect3 or an (attrui)
  4. Start dp.kinect3, the crash/freeze usually happens within a few seconds.
* Microsoft will not fix bug <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1540>.
  When you use body tracking, the model file must be on an ASCII path.
* The second parameter of `@skelcompute directml 0` is the numeric index of the DirectML
  device used for body tracking. This index may change because the Windows Graphics Performance
  Preference default is "Let Windows decide". Windows may decide to change the order of GPUs.
  This index *is* consistent when you choose a Graphics Performance Preference. This
  setting is in Windows settings, System, Display, Graphics settings.
* Enabling `@opencl` on an Intel GPU may crash Max when you first start your sensor. This
  is a known bug with Intel GPU drivers and LLVM (the compiler Intel uses for OpenCL).
  The only workaround is to try again. After clang successfully compiles the OpenCL kernels,
  they will be cached on-disk by OpenCV. These on-disk cached kernels will be used the next
  time and the LLVM clang compiler will not be used; therefore avoiding the crash. This cache
  is not forever and will be purged due to disk usage, Intel driver update, etc. See
  <https://github.com/llvm/llvm-project/issues/29935>, <https://github.com/intel/compute-runtime/issues/122>,
  <https://github.com/NixOS/nixpkgs/issues/97401>.
  For reference, these are the errors that appear in a debugger console during a crash.
  ```
  Loaded 'C:\Windows\System32\DriverStore\FileRepository\iigd_dch.inf_amd64_afa18c61a36f2728\opencl-clang64.dll'. Symbol loading disabled by Include/Exclude setting.
  clang: CommandLine Error: Option 'h' registered more than once!
  LLVM ERROR: inconsistency in registered CommandLine options
  Exception thrown at 0x00007FFC418A79A4 (opencl-clang64.dll) in Max.exe: 0xC0000005: Access violation reading location 0x0000000000000000.
  Exception thrown at 0x00007FFC418A79A4 (opencl-clang64.dll) in Max.exe: 0xC0000005: Access violation reading location 0x0000000000000000.
  Exception thrown at 0x00007FFC418A79A4 (opencl-clang64.dll) in Max.exe: 0xC0000005: Access violation reading location 0x0000000000000000.
  Exception thrown at 0x00007FFC418A79A4 (opencl-clang64.dll) in Max.exe: 0xC0000005: Access violation reading location 0x0000000000000000.
  The program '[665836] Max.exe' has exited with code 1 (0x1).
  ```

## v1.3.20220321

* Improved frame synchronization
* Registration problems logged to %localappdata%\Dale Phurrough\dp.kinect3
  for local diagnostics
* Improve error reporting to Max console

## v1.3.20211227

* No feature changes
* Updates of 3rd party libraries

## v1.3.20210818 (first beta release)

### New features

* Joint orientations in all four formats now work
* jit.anim.node transform hierarchies are supported
* accelerometer and gravity data; rotate to gravity `@rotatemethod=2`
* Add ability to select the OpenCL device using the `opencl` attribute.
  This and other compute choices `decodercolor` and `skelcompute` allow
  you to better utilize modern computing devices.
  ```
  @opencl
           0     off
           1     default OpenCL device or overridden with OPENCV_OPENCL_DEVICE
                 environment variable; see usage in this doc or OpenCV docs
           2...  specific OpenCL device
  ```
* Performance work to reduce CPU and latency; improved error handling

### Changes

* Changed order of joint output to reverse hierarchy. Previously `waist` was first.
  The new order clarifies hierarchy and facilitates puppet animation.
  ```
  l_thumb, l_hand_tip, l_hand, l_wrist, l_elbow, l_shoulder,
  r_thumb, r_hand_tip, r_hand, r_wrist, r_elbow, r_shoulder,
  l_foot, l_ankle, l_knee, l_hip,
  r_foot, r_ankle, r_knee, r_hip,
  head, neck, c_shoulder, torso, waist
  ```
* `accel` attribute is now 3 floats (x,y,z). Previously it was four (x,y,z,0)
* Removed deprecated `elevation` attribute. Use `accel` or `gravrot` for the
  current rotation of the sensor relative to gravity.
* Removed deprecated `@rotatemethod` method `1=elevation`. Methods `0` and `2`
  are supported. Use `rotate`, `rotatexyz`, or `quat` for X axis rotation.
* Deprecated `idkinect`. It will be removed in 12-18 months. Use `idsensor`.

### Known issues

* `flipx` and `rotatemethod` can not be used at the same time
* OpenCV issue <https://github.com/opencv/opencv/issues/20486> will cause a small
  device memory leak when changing between OpenCL devices. There is no known
  memory leak when you use the same device. For example
  ```
  off -> Intel  -> off -> Intel...     no memory leak
  off -> NVidia -> off -> NVidia...    no memory leak
  off -> Intel  -> ... -> NVidia...    memory leak 😞
  ```

## v1.3.20210721

* Joint orientations in absolute world coordinates now work
* Updated some joint positions to be similar to dp.kinect and dp.kinect2
* Add align of all data to depth resolution/coordinates `@align 1`
* Add depth visibility threshold to eliminate double images `@depthvis` with
  the same default value and behavior as dp.kinect2
* Add filter of data by playermap with `@depthonlyplayers`
* Some attributes can now be set by symbols
  ```
  @align color
  @depthonlyplayers color+ir+depth
  @rotatemethod gravity
  @skeletonformat max
  ```
* Add OSC format skeleton output `@skeletonformat 0`
* Performance and memory improvements

## v1.3.20210701

* Add dp.kinect3 helpfile. It is a copy of the dp.kinect2 helpfile with
  `2` changed to `3`. Some examples work. Others need features not yet implemented
* Joint orientations for `waist`, `torso`, `head`, and `neck` are *probably* correct.
  Other joint orientations are not correct. Work is ongoing to improve these.
* Fix startup bug when sensor might not output data or crash
* Minor code profiling/perf improvements

## v1.3.20210609

* Add `@smoothing`. The default `0.0` is no smoothing. The attribute will also
  accept the 5-parameter smoothing from dp.kinect and dp.kinect2 -- however it will
  only act on the 1st parameter. dp.kinect3 will soon support all five parameters
  allowing you to use the short 1-param or the full 5-params.
* Internal code refactoring
* Update some third-party libraries

## v1.3.20210520

* Peliminary joint orientation in absolute formats; not yet reliable
* Internal code refactoring
* Build pipeline improvements and additional testing tools

## v1.3.20210429

* Body tracking `tensor` and `tensor_fp16` runtimes have a slow memory leak.
  It is a problem in the body tracking components from Microsoft.
  <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1576>
  Every 45 minutes: approximately 1GB is consumed, 1-2 fps slower, and a matching lag.
  The other body tracking `@skelcompute` choices `cpu`, `cuda`, and `directml`
  do not have this memory leak.
* `@colormap` symbolic names; `@colormap argb` and `@colormap 1` both work
* Minor performance improvements
* Code restructuring and update of 3rd party libraries

## v1.3.20210409

* Upgrade to Microsoft Body Tracking v1.1.0 enables more GPUs (Intel, AMD, NVIDIA, etc.) and
  performance tuning opportunities
* `@skelcompute` supports two arguments: compute type and device index. The default
  is `directml` and device index `0`.
* Optimize pointcloud generation; often 4 to 10x faster
* Optimize OpenCL memory to reduce latency

#### Body tracking details

* Body Tracking install has changed. Follow the instructions in `INSTALL.md`.
  The number of Body Tracking DLLs and file sizes are both **substantially larger**.
* Microsoft provides two body tracking models; the default `dnn_model_2_0_lite_op11.onnx` and
  `dnn_model_2_0_op11.onnx`. The non-lite model is slightly more accurate and uses more resources.
* NVIDIA gpus often perform better with TensorRT when compared to CUDA. I recommend testing all
  compute types (cuda, directml, tensor, etc.).
* `@skelcompute tensor_fp16` will use the Tensor Runtime and optimize the chosen body tracking model
  with 16-bit floating point calculations. It is slightly less accurate and uses less resources.
* TensorRT choices will have a **very long** first-time startup. For example, it is almost 5 minutes on
  my laptop with a RTX2070-Super gpu when I choose `tensor_fp16`. Later startups are only a few seconds.
* TensorRT caches the first-time startup optimizations at `%TEMP%\dp.kinect3\tensor.cache`.
  You or Windows can delete these cache folders. TensorRT will re-create a cache
  on the next startup.
* Each TensorRT cache folder is dependent on your Body Tracking DLLs, the body tracking model,
  the gpu, and fp16 optimization. If you change any of those, a new cache folder will be created and
  new optimizations cached. These cache folders can grow to be hundreds of megabytes.

#### @skelcompute examples

```
@skelcompute directml       <-- DirectML with the first gpu that is DirectML capable
@skelcompute directml 1     <-- DirectML with the second gpu that is DirectML capable
@skelcompute cuda           <-- CUDA with the first NVIDIA gpu
@skelcompute tensor         <-- TensorRT with the first NVIDIA gpu
@skelcompute tensor_fp16    <-- TensorRT with the first NVIDIA gpu, 16-bit floating point
```

## v1.3.20210320

This release focuses on a new attribute `timestamp`. This attribute will add
timestamps to skeleton and matrix data. These timestamps are microseconds
since the Epoch. <https://en.wikipedia.org/wiki/Unix_time>

Jitter matrix outlets (depth, color, ir, player, and optional pointcloud) will
append the timestamp to the end of the standard `jit_matrix` message.

```
jit_matrix u391000480                   <-- standard matrix message

jit_matrix u391000480 1616264945444860  <-- same with timestamp
```

Message/OSC data is also timestamped. A `timestamp` message sent on the
msg/osc outlet starts a timestamp block. All data after this timestamp message and
before the next timestamp message have the same timestamp.

```
timestamp 1616264778766184
user 1 0.092804 -0.510639 0.794896
skel 1 waist 0.092804 -0.510639 0.794896 1.
skel 1 torso 0.113126 -0.20709 0.847535 1.
...
```

It is possible to coordinate matrices and skeleton data across multiple
Kinect sensors on the same computer using these timestamps. If you have
*very accurate* NTP or PTP time synchronization on your local network,
it is also possible to coordinate data across multiple computers.

## v1.3.20210226

* Add `powerfreq`, `indicator`, and color `exposure` control; the latter
  is important when synchronizing multiple sensors
  <https://docs.microsoft.com/en-us/azure/kinect-dk/multi-camera-sync#exposure-considerations>
* Workaround two Azure Kinect SDK issues
* Add and update text warnings

## v1.3.20210224

* Add synchronization settings to coodinate multiple Kinects with
  `depthdelay`, `syncdelay`, `syncmode`, and `idkinect`
* Add `getports` to retrieve status of sync-out and sync-in ports
* Add `getusbidlist` to retrieve ids for each connected Kinect
* Workaround Azure Kinect SDK and Intel Media SDK bugs
* Change default fps to 30fps for backwards compatibility
* Resolve all known issues in libyuv decoder
* Optimize IR processing
* Ongoing optimization of data pipeline

## v1.3.20210207

* substantial improvements for more asynchronous/parallel processing
* fix several opencl, skeleton, and shutdown bugs

## v1.3.20210130

* performance improvements
* add uyvy flipx
* update Windows SDK
* add install doc for NVIDIA color decoder

## v1.3.20210111

* Add NVIDIA GPU color decoder
* uyvy support across all color decoders; no uyvy postprocessing
* Performance improvements to other color decoders
* Update text/error messages

## v1.3.20210105

* Significantly better Kinect data processing reliability
* Workaround for OpenCV memory leaks
* Changed default pipeline method to `async`
* Fixed data sync on all outlets except skeleton joints
* Add very limited (1280x720) uyvy output
* Add checks for dependent DLLs

## v1.3.20201228

* More work on multi-core/threads to process camera data into jitter matrices
* Known issue: data from each outlet is not synchronized

## v1.3.20201127

### Feature changes

* First stage of new multi-threaded approach to process sensor data
* Remove Mango color decoder; too much thread activity when not decoding

### Performance

* Better use of multiple CPU cores

### Fixed

* Use of multiple compute device for different tasks using the
  attributes @decodercolor, @opencl, and @skelcompute. For example,
  you could...
  * decode color frames on the video decoder in Intel CPU
  * flip and undistort frames on integrated Intel GPU
  * detect skeleton joints on the discrete Nvidia GPU
  * and the remaining features run on your CPU
* Workaround dependent library shutdown issues

## v1.3.20201027

### Feature changes

* Add support for some backward-compatible resolutions
* Change default depthmap to be narrow field of view
* Improve infrared output quality
* Add Windows Media Foundation color decoder; made default decoder

### Performance

* Optimize infrared gamma adjustments
* Optimize joint calculation

### Fixed

* align skeleton joints to color coordinates when @align=color
* example patch pointcloud frame sync

## v1.3.20201007

### Feature changes

* Enable playermap output
* Align, undistort, and flipx works with more output

### Performance

* More utilization of GPU/CPU in body tracking neural net

### Fixed

* Alignment and flipx bugs and crashes
* OpenCL crashes
* Synchronization of depth, color, ir, player, and skelton data

## v1.3.20200918

No significant changes in this release. The next release will include updates.

## v1.3.20200905

### Feature changes

* Body joint tracking with position and confidence with `@skeleton`.
* New attribute `@skelcompute` allows you to choose your GPU or CPU
  to run the tracking algorithm. Please read below about hardware requirements.
* Azure Kinect 3 does not have a maximum number of bodies it can track.
  The `playerID` value assigned to a tracked body can vary widely
  if you are tracking many bodies. The IDs will be recycled and the set of
  active IDs will tend towards the value `1`.
* Support alignment of depth/ir/pointcloud to the chosen color resolution with `@align`
* Ongoing refactoring and bugfixes

### Hardware Requirements

Microsoft has designed the Azure Kinect 3 to be used with GPUs and neural networks.
It is possible to use only a CPU, but your performance will be significantly lower.
Please verify your computer meets Microsoft's system requirements <https://docs.microsoft.com/en-us/azure/kinect-dk/system-requirements>

Courtesy repeat 📣  
When using body tracking...  
...and you want to run body tracking on a _GPU_  
...your GPU must be an 🚨Nvidia GPU🚨.

Microsoft is aware of the Nvidia limitation. They have not announced any change.

Microsoft is also aware of the performance of body tracking <https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/514>. Microsoft's official statement is,

> ...[in the next semester 2020H2] we are focusing on a more performant DNN.

High-framerate body tracking _today needs strong hardware_.
Naturally, your specific body tracking needs determines your hardware.

I have explored other body tracking technology but have not yet found a solution that performs well and has a licensing model that works. One company is reworking their licensing model and I'm waiting to see the outcome. I will keep looking :-)

## v1.3.20200813

### Feature changes

* Limited features are implemented in this release. The final
  release will support almost all features of dp.kinect2;
  high detail face tracking may not be possible.
* Depth, color, infrared, and pointcloud outputs are available. Aligning
  them is not yet available.
* You can undistort the depth, color, and infrared output. Naturally, there
  is a small CPU cost for this. Use the `@undistort` and `@undistortscale`
  attributes to control this.
* Microsoft changed the resolutions for the sensor. Therefore,
  resolutions on dp.kinect3 are chosen by specifying the resolution
  width and height -- not an index. For example, if you want the
  full HD colormap, then specify `@colormapres 1920x1080`. You can use
  any character between the width and height except for a period `.`  
  Color default is `1280x720`  
  Depth and infrared defaults are `512x512`
* You can choose the frames per second for the sensor using the
  `@depthfps` attribute

### Performance, Acceleration, and GPU

* Microsoft does not provide low-latency full HD (1920x1080) color output. Instead,
  they only provide one low-latency color output at base HD (1280x720) resolution.
  If you need low-latency color output, I recommend you use 1280x720 resolution.
* Many higher resolutions of color output are available (including 4K). However,
  these resolutions are only captured by the sensor in MJPEG format. MJPEG is
  an outdated compression format; an unfortunate Microsoft choice. Due to the
  mathematics of MJPEG, the decompression can not be low-latency multithreaded on the CPU or GPU.
  This means -- on modern CPUs with multiple "cores", only one core will be used
  to decompress this MJPEG data. When that one core is running as fast as possible,
  then color frames can not be processed any faster. You may see this in
  Windows 10 Task Manager; one core may be at 100% while other cores are not busy.
* Due to this limitation with high-resolution MJPEG color output, I have
  included three (3) different color decompressors with this alpha release.
  You can choose any of the three using the `@decodercolor` attribute. The value
  can be `libyuv`, `intelmedia` or `mango`. The default is `@decodercolor libyuv`.
  * `libyuv` selects the open source libyuv decompression library. This runs only
    on the CPU and is not multithreaded. This is the default decompressor.
  * `intelmedia` selects the Intel Media SDK decompressor. If you have an Intel CPU
    with Intel Quick Sync/Graphics, then this decoder will use the hardware Quick
    Sync accelerator.
  * `mango` selects the open source MANGO Multimedia Development Platform
    by Twilight Finland 3D Oy Ltd. It is able to do *some stages of MJPEG decompressing
    in parallel*. This decompressor usually enables higher framerates at the cost
    of more resource usage.
* Some image processing (e.g. flipx, undistort, align, etc.) may be done on your
  GPU or accelerator if your drivers support OpenCL v1.2. Your peformance may be
  better or worse. It depends on your specific hardware and the specific image
  processing you choose. The default is off (do not use OpenCL). You can
  enable `@opencl 1` or disable `@opencl 0`. OpenCV is used to implement this
  feature and OpenCV will automatically chose your default OpenCL device when
  you enable OpenCL.
  * You can see this default OpenCL device when you look at the `@opencl`
    attribute with an (attrui) or the Max Inspector and look at the description at index 1.
    You will see all your compatible OpenCL devices listed, but you can only select
    `0=disable` or `1=default`.
  * You *can* change this OpenCL default by setting the `OPENCV_OPENCL_DEVICE`
    environment variable
    <https://github.com/opencv/opencv/wiki/OpenCL-optimizations#opencv-opencl-configuration-options>
    ```
    <Platform>:<CPU|GPU|ACCELERATOR|nothing=GPU/CPU>:<DeviceName or ID>

    Examples:
        :GPU:1
        AMD::Tahiti
        Intel:CPU:
        NVIDIA:GPU:0
    ```
