# dp.kinect3

dp.kinect3 is a plugin for the [Cycling '74 Max](https://cycling74.com/) development environment to use your [Microsoft Azure Kinect](https://azure.microsoft.com/en-us/services/kinect-dk/) sensor (the 3rd generation Kinect) on your Windows PC.

Microsoft has failed to provide reliable and consistent updates to their drivers and software.
You may encounter minor faults in dp.kinect3. Our attempts to improve are hindered by
Microsoft's lack of progress and communication.
<https://github.com/microsoft/Azure-Kinect-Sensor-SDK/issues/1824>

Luxonis is an better sensor alternative...

[Luxonis](https://shop.luxonis.com/) has a wide variety of depth and AI sensors.
They demonstrate the ability to deliver, iterate and improve, and partner with the community.
We have a plugin that works with these sensors. You can learn more at <https://hidale.com/shop/dp.oak>.

### Hardware and Download Requirements 

* [Cycling74 Max requirements](https://cycling74.com/products/max)
* [Microsoft's Kinect requirements](https://docs.microsoft.com/en-us/azure/kinect-dk/system-requirements)
* Windows 10 April 2018 (Version 1803, OS Build 17134) release (64-bit) or newer

The Azure Kinect sensor does not have a shared driver from Microsoft that all applications use.
This approach by Microsoft requires customers to install their own files. There are three main groups:

* dp.kinect3 files that we create and test. Most features _except body tracking_  work
  with only these files. ~19 MB download
* Microsoft body tracking files that Microsoft creates and tests. ~430 MB download

### Setup

1. Install [Cycling'74 Max](https://cycling74.com/downloads/) version 6.1.9 or newer (64-bit)
2. Download dp.kinect3 from <https://hidale.com/shop/dp-kinect3/>
3. [Decompress](https://support.microsoft.com/en-us/help/14200/windows-compress-uncompress-zip-files)
   all the files in this download into an empty directory for dp.kinect3
4. Next, you must register dp.kinect3

### Register

1. You can evaluate dp.kinect3 with a two week free trial. If you want to use
   it after trying it, please visit <https://hidale.com/shop/dp-kinect3/> to purchase a license.
2. Open the .maxhelp file included within the ZIP download.
3. Click the top-left `Register` tab.
4. Type your registration name in the field beside the register button. You
   received your registration name in an email from our store. Hint: your registration
   name is usually _not your email address_.
5. Click the register button and use the dialog box that appears to select
   your dp.kinect3 registration key (.dpreg file)
6. You will see a successful registration in the Max console window

### Body Tracking Setup

1. Download and install Microsoft's
   [Body Tracking SDK 1.1.2](https://docs.microsoft.com/en-us/azure/kinect-dk/body-sdk-download).
   Note the install directory; often `"C:\Program Files\Azure Kinect Body Tracking SDK"`
2. Navigate into the SDK install directory, then the `tools` subdirectory,
   e.g. `"C:\Program Files\Azure Kinect Body Tracking SDK\tools"`
3. Copy these files from `sdk\windows-desktop\amd64\release\bin` to your dp.kinect3 directory
   ```
   k4abt.dll
   dnn_model_2_0_lite_op11.onnx
   dnn_model_2_0_op11.onnx      <- slightly more accurate
   ```

### Documentation

New documentation and tutorials are at <https://docs.hidale.com/>.

Existing patches may work with only a change of the object name to (dp.kinect3).

### Advanced GPU and Computing Setup

The plugin has settings for compute and GPU devices

* `@opencl` for matrix processing -- default is disabled
* `@skelcompute` for body tracking neural net -- default is primary GPU
* `@transcoder` for color camera decoding -- default is Windows Media Foundation

We recommend you experiment to discover which settings meet your needs for hardware,
latency, and throughput. You can have significant performance improvements! 🙂

#### OpenCL computation

Some image processing may be done on your GPU or accelerator if your device and drivers
support OpenCL v1.2. Your peformance may be better or worse; depends on specific
hardware and image processing you choose. In our testing, integrated GPUs often perform well.

OpenCL is disabled by default. You can enable `@opencl 1` or disable `@opencl 0`.
The plugin will use your default OpenCL device when you enable OpenCL. View all OpenCL devices
at the `@opencl` attribute using an (attrui) or the Max Inspector.

#### Body Tracking computation

`@skelcompute` supports two arguments: compute type and device index.
The default is `directml` and device index `0`. We recommend DirectML.
DirectML works on all modern GPUs and performs the same or better
during testing than other compute types.

```
@skelcompute directml        <- DirectML with first gpu
@skelcompute directml 1      <- DirectML with second gpu
@skelcompute directml intel  <- DirectML with gpu named intel
@skelcompute cuda            <- CUDA with the first NVIDIA gpu
@skelcompute tensor          <- TensorRT with the first NVIDIA gpu
@skelcompute tensor_fp16     <- TensorRT with the first NVIDIA gpu, 16-bit floating point
```

##### Optional NVIDIA CUDA and TensorRT for Body Tracking

NVIDIA GPUs work well with the DirectML support built into the plugin.
It is possible to use optional CUDA and TensorRT providers for unusual situations.

NOTE: During testing, these optional CUDA and TensorRT providers
performed the same or _worse_ than DirectML.

The setup requires ~4 GB of additional downloads and the setup is documented at
<https://github.com/diablodale/dp.kinect3/wiki#optional-nvidia-only-body-tracking-with-cuda-and-tensorrt>.

NVIDIA gpus often perform better with TensorRT when compared to CUDA. Please
note the following

* `@skelcompute tensor_fp16` will use the Tensor Runtime and optimize the chosen body tracking model
  with 16-bit floating point calculations. It is slightly less accurate while using less resources.
* TensorRT choices will have a **very long** first-time startup. For example, it is almost 5 minutes on
  our test laptop with a RTX2070-Super gpu when we choose `tensor_fp16`. Later startups are only a few seconds.
* TensorRT caches the first-time startup optimizations at `%TEMP%\dp.kinect3\tensor.cache`.
  You or Windows can delete these cache folders. TensorRT will re-create a cache
  on the next startup.
* Each TensorRT cache folder is dependent on your Body Tracking DLLs, the body tracking model,
  the gpu, and fp16 optimization. If you change any of those, a new cache folder will be created and
  new optimizations cached. These cache folders can grow to be hundreds of megabytes.
